# Copyright (C) 2004 Philippe Normand <phil@respyre.org>
#
# This file is part of XMLObject (http://xmlobject.base-art.net)
#
# XMLObject is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# ResPyre is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ResPyre; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# xml.minidom is only used to parse incoming XML
# when building XMLObject from string data

from xml.dom.minidom import parseString

import classregistry

class MetaAttribute(object):
    """ Each XMLObject type deals with MetaAttributes. In the example below,
        `content` and `title` are MetaAttributes::

          >>> class MyXMLObject(XMLObject):
          ...   content = TextNode('Content') # _XO_content
          ...   title = StringAttribute()     # _XO_title
          ...
          >>>
          
        MetaAttributes handle value with the XMLObject they're included in.
    """
    
    def __init__(self, nodeInstance):
        self._name = '_XO_%s' % nodeInstance.getName()
        self._node = nodeInstance
        
    def __get__(self, inst, cls):
        return inst.get(self._name)

    def __set__(self, inst, val):
        if self._node.checkType(val):
            inst.set(self._name,val)
        else:
            raise TypeError('incorrect type in assignment of '+ `self._node.getName()`
                            + ' (%s)' % str(val))
        
class MetaXMLObject(type):
    """ The place where all magic happens:

        - remember the name of the new type
        - store attribute instances (used in XMLObject.{to,from}Xml methods)
        - create MetaAttributes
        - handle inheritance

        Special variables are:

        - `__name__` : the class name
        - `__nodes__` : the hash of class members keyed by name. Each node has
          to be an instance of the XONode class.
    """

    # XXX: may introspection do better work here ?
    __methods__ = [ '_isMember', 'set','get',
                    'toXml', 'fromXml', '_fromDom','getName']

    def __new__(cls, className, bases, dictionnary):
        dictionnary['__name__'] = className
        dictionnary['__nodes__'] = {}
        for name,instance in dictionnary.items():
            if '_' not in name[:1] and name not in cls.__methods__:
                # store the attribute name somewhere
                instance.setName(name)
                if instance.getType() in ['List','Item']:
                    instance.setParentType(className)
                dictionnary['__nodes__'][name] = instance
                dictionnary[name] = MetaAttribute(instance)

        newClass = type.__new__(cls, className, bases, dictionnary)

        # inheritance handling
        superClasses = cls.mro(newClass)[1:-2]
        if len(superClasses) > 0:
            newClass.__nodes__ = {}
        for superClass in superClasses:
            for nodeName,node in superClass.__nodes__.items():
                newClass.__nodes__[nodeName] = node
                setattr(newClass, nodeName, MetaAttribute(node))

        # finally register the new class type so that it can be found
        # using its name
        classregistry.registry(newClass._registry).addClass(newClass)
        return newClass

class XMLObject:
    """ 

    """
    
    __metaclass__ = MetaXMLObject

    _registry = None

    
    def __init__(self, **kw):
        self._attributes = {}

        # set default values for the new class instance
        for attrName, instance in self.__nodes__.items():
            #if not instance.isFacultative():
            setattr(self,attrName,instance.getValue())

        # set user values for the class instance
        # Some may override defaults
        for k,v in kw.items():
            if self._isMember(k):
                setattr(self, k, v)

    def _isMember(self, name):
        " Dummy method to drop later "
        result = filter(lambda i: i == name, self.__nodes__.keys())
        return len(result) > 0

    def getName(self):
        return self.__name__

    def get(self, nodeName):
        return self._attributes[nodeName]

    def set(self, nodeName, val):
        self._attributes[nodeName] = val
        
    def toXml(self, indent = 0, headers = 1, tabLength=4, encoding="iso-8859-1"):
        result = ''
        if headers:
            result = '<?xml version="1.0" encoding="%s" ?>\n' % encoding
        attrs, others = '', ''
        currIndent = ' ' * indent
        result += currIndent 
        result += '<%s ' % self.getName()
        for attrName,node in self.__nodes__.items():
            # backup default node value
            default = node.getValue()
            # set node value temporary
            node.setValue(getattr(self, attrName))
            if node.getType() in ['StringAttribute', 'IntegerAttribute']:
                attrs += node.xmlrepr() + ' '
            else:
                others += node.xmlrepr(indent=indent+tabLength)
            # reset value to default
            node.setValue(default)
        result += attrs
        if others != []:
            result = '%s>\n' % result[:-1]
            result += others
            result += currIndent + '</%s>' % self.getName()
        else:
            result = '%s/>' % result[:-1]
        return result

    def fromXml(self, xmlData):
        dom = parseString(xmlData)
        self._fromDom(dom.documentElement)
        dom.unlink()

    def _fromDom(self, dom):
        for attrName, node in self.__nodes__.items():
            try:
                value = node.getValueFromDom(dom, attrName, registry=self._registry)
            except Exception, e:
                raise
            else:
                setattr(self, attrName, value)
        return self

__all__ = [ 'XMLObject' ]
